﻿using HeuristicLab.Operators;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Analysis;

namespace HeuristicLab.Algorithms.SimulatedAnnealing
{
    [Item("ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator", "The operator resets the temperature to X when N consecutive solutions are rejected. X is calculated like the initial temperature based on average hillsize.")]
    [StorableClass]
    public class ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator : SingleSuccessorOperator, IReheatingOperator
    {
        #region Strings
        private const string AnnealingOperatorName = "AnnealingOperator";
        private const string LowerTemperatureName = "LowerTemperature";
        private const string IterationsName = "Iterations";
        private const string TemperatureStartIndexName = "TemperatureStartIndex";
        private const string StartTemperatureName = "StartTemperature";
        private const string EndTemperatureName = "EndTemperature";
        private const string TemperatureName = "Temperature";
        private const string MaximumIterationsName = "MaximumIterations";
        private const string InitialTemperatureName = "InitialTemperature";

        private const string ThresholdName = "Threshold";
        private const string ReheatRatioName = "ReheatRatio";
        private const string IsAcceptedName = "IsAccepted";

        private const string ConsecutiveRejectedSolutionsCountName = "ConsecutiveRejectedSolutions";
        private const string LastQualityName = "LastQuality";
        private const string UphillMovesMemoryName = "UphillMovesMemory";
        private const string MaximizationName = "Maximization";
        private const string MoveQualityName = "MoveQuality";
        private const string InitialAcceptanceRateName = "InitialAcceptanceRate";
        #endregion

        #region Parameters
        public ILookupParameter<DoubleValue> TemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[TemperatureName]; }
        }
        public ILookupParameter<DoubleValue> LowerTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[LowerTemperatureName]; }
        }
        public ILookupParameter<DoubleValue> StartTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[StartTemperatureName]; }
        }
        public ILookupParameter<DoubleValue> EndTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[EndTemperatureName]; }
        }
        public ILookupParameter<IntValue> TemperatureStartIndexParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[TemperatureStartIndexName]; }
        }
        public ILookupParameter<IntValue> IterationsParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[IterationsName]; }
        }
        public ILookupParameter<IOperator> AnnealingOperatorParameter
        {
            get { return (ILookupParameter<IOperator>)Parameters[AnnealingOperatorName]; }
        }
        public ILookupParameter<IntValue> MaximumIterationsParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[MaximumIterationsName]; }
        }
        private ValueParameter<IntValue> ThresholdParameter
        {
            get { return (ValueParameter<IntValue>)Parameters[ThresholdName]; }
        }
        private ValueParameter<DoubleValue> ReheatRatioParameter
        {
            get { return (ValueParameter<DoubleValue>)Parameters[ReheatRatioName]; }
        }
        public ILookupParameter<BoolValue> IsAcceptedParameter
        {
            get { return (ILookupParameter<BoolValue>)Parameters[IsAcceptedName]; }
        }
        public ILookupParameter<IntValue> ConsecutiveRejectedSolutionsCountParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[ConsecutiveRejectedSolutionsCountName]; }
        }
        public ILookupParameter<ItemList<DoubleValue>> UphillMovesMemoryParameter
        {
            get { return (ILookupParameter<ItemList<DoubleValue>>)Parameters[UphillMovesMemoryName]; }
        }
        public ILookupParameter<DoubleValue> MoveQualityParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[MoveQualityName]; }
        }
        public ILookupParameter<DoubleValue> LastQualityParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[LastQualityName]; }
        }
        public IValueLookupParameter<BoolValue> MaximizationParameter
        {
            get { return (IValueLookupParameter<BoolValue>)Parameters[MaximizationName]; }
        }
        public ILookupParameter<DoubleValue> InitialAcceptanceRateParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[InitialAcceptanceRateName]; }
        }
        #endregion

        public ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator() : base()
        {
            #region Create parameters
            Parameters.Add(new LookupParameter<DoubleValue>(TemperatureName, "The current temperature."));
            Parameters.Add(new LookupParameter<DoubleValue>(LowerTemperatureName, "The lower bound of the temperature."));
            Parameters.Add(new LookupParameter<IntValue>(IterationsName, "The number of iterations."));
            Parameters.Add(new LookupParameter<IOperator>(AnnealingOperatorName, "The operator that cools the temperature."));
            Parameters.Add(new LookupParameter<IntValue>(TemperatureStartIndexName, "The index where the annealing or heating was last changed."));
            Parameters.Add(new LookupParameter<DoubleValue>(StartTemperatureName, "The temperature from which cooling should occur."));
            Parameters.Add(new LookupParameter<DoubleValue>(EndTemperatureName, "The temperature to which should be cooled."));
            Parameters.Add(new LookupParameter<IntValue>(MaximumIterationsName, "The maximum number of iterations which should be processed."));
            Parameters.Add(new LookupParameter<BoolValue>(IsAcceptedName, "Whether the move was accepted or not."));
            Parameters.Add(new LookupParameter<IntValue>(ConsecutiveRejectedSolutionsCountName, "Amount of consecutive rejected solutions."));
            Parameters.Add(new ValueParameter<IntValue>(ThresholdName, "How many consecutive rejected solutions must occur to trigger a reheat.", new IntValue(10)));
            Parameters.Add(new ValueParameter<DoubleValue>(ReheatRatioName, "The acceptance rate of average-sized hills used to calculate the reheat temperature.", new DoubleValue(0.4d)));

            Parameters.Add(new ValueLookupParameter<BoolValue>(MaximizationName, "True if the problem is a maximization problem, otherwise false."));
            Parameters.Add(new LookupParameter<ItemList<DoubleValue>>(UphillMovesMemoryName, "Memorizes the last 100 uphill moves."));
            Parameters.Add(new LookupParameter<DoubleValue>(MoveQualityName, "The value which represents the quality of a move."));
            Parameters.Add(new LookupParameter<DoubleValue>(LastQualityName, "The value which represents the quality of the last move."));
            Parameters.Add(new LookupParameter<DoubleValue>(InitialAcceptanceRateName, "The initial acceptance rate of average-sized hills used to calculate the initial temperature."));

            #endregion

            Parameterize();
        }

        [StorableConstructor]
        protected ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator(bool deserializing) : base(deserializing) { }
        protected ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator(ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator original, Cloner cloner) : base(original, cloner) { }


        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new ConsecutiveRejectionTemperatureResetBasedOnAverageHillsizeOperator(this, cloner);
        }

        public void Parameterize()
        {
        }

        public override IOperation Apply()
        {
            var currentQuality = MoveQualityParameter.ActualValue.Value;
            var lastQuality = LastQualityParameter.ActualValue.Value;
            var isMaximation = MaximizationParameter.ActualValue.Value;
            var hillMemory = UphillMovesMemoryParameter.ActualValue;

            var currentHillSize = CalculateHillSize(lastQuality, currentQuality, isMaximation);

            if (currentHillSize > 0)
            {
                hillMemory.Add(new DoubleValue(currentHillSize));
                if (hillMemory.Count > 100) hillMemory.RemoveAt(0);
            }

            var count = ConsecutiveRejectedSolutionsCountParameter.ActualValue;

            count.Value = !IsAcceptedParameter.ActualValue.Value ? count.Value + 1 : 0;

            var heating = count.Value == ThresholdParameter.Value.Value;

            if (heating) count.Value = 0;

            return heating ? Heat(hillMemory) : Cool();
        }

        private IOperation Heat(ItemList<DoubleValue> hillMemory)
        {
            var temperature = Math.Max(CalculateTemperatureBasedOnAverageHillSize(hillMemory), TemperatureParameter.ActualValue.Value);
            TemperatureParameter.ActualValue.Value = temperature;
            TemperatureStartIndexParameter.ActualValue.Value = Math.Max(0, IterationsParameter.ActualValue.Value - 1);
            StartTemperatureParameter.ActualValue.Value = TemperatureParameter.ActualValue.Value;
            return base.Apply();
        }

        private IOperation Cool()
        {
            return new OperationCollection {
                    ExecutionContext.CreateOperation(AnnealingOperatorParameter.ActualValue),
                    base.Apply()
                };
        }

        private double CalculateHillSize(double lastQuality, double currentQuality, bool isMaximation)
        {
            return lastQuality < 0 ? -1 : isMaximation ? lastQuality - currentQuality : currentQuality - lastQuality;
        }

        private double CalculateTemperatureBasedOnAverageHillSize(ItemList<DoubleValue> hillMemory)
        {
            var averageHillSize = hillMemory.Average(x => x.Value);
            var medianHillsize = hillMemory.OrderByDescending(x => x.Value).ElementAt(hillMemory.Count / 2 - 1).Value;
            var avgtemp = -(averageHillSize / Math.Log(ReheatRatioParameter.Value.Value));
            var mediantemp = -(medianHillsize / Math.Log(ReheatRatioParameter.Value.Value));
            Console.Out.WriteLine("AVG:    " + averageHillSize + " " + avgtemp);
            Console.Out.WriteLine("MEDIAN: " + medianHillsize + " " + mediantemp);
            Console.WriteLine();

            return mediantemp;
        }
    }
}
